#!/usr/bin/env python

import numpy as np
import sklearn.metrics
import sklearn.metrics.pairwise
from scipy.stats.stats import pearsonr  
from sklearn.metrics.cluster import normalized_mutual_info_score
import matplotlib.pyplot as plt; plt.rcdefaults()
import matplotlib.pyplot as plt
from sklearn.preprocessing import KBinsDiscretizer
from sklearn.metrics import accuracy_score
from wuml.opt_gaussian import *
import pandas as pd
import numpy as np
import ppscore as pps

# compute normalized HSIC between X,Y
# if sigma_type = mpd, it uses median of pairwise distance
# if sigma_type = opt, it uses optimal
def HSIC(X,Y, X_kernel='Gaussian', Y_kernel='Gaussian', sigma_type='opt', normalize_hsic=True):	
	def get_γ(X,Y, sigma_type):
		if sigma_type == 'mpd': 
			σᵪ = np.median(sklearn.metrics.pairwise_distances(X))		
			σᵧ = np.median(sklearn.metrics.pairwise_distances(Y))
		else: 
			optimizer = opt_gaussian(X,Y, Y_kernel=Y_kernel)
			optimizer.minimize_H()
			σᵪ = optimizer.result.x[0]
			σᵧ = optimizer.result.x[1]
			if σᵪ < 0.01: σᵪ = 0.01		# ensure that σ is not too low
			if σᵧ < 0.01: σᵧ = 0.01		# ensure that σ is not too low

		γᵪ = 1.0/(2*σᵪ*σᵪ)
		γᵧ = 1.0/(2*σᵧ*σᵧ)

		return [γᵪ, γᵧ]

	def double_center(Ψ):
		HΨ = Ψ - np.mean(Ψ, axis=0)								# equivalent to Γ = Ⲏ.dot(Kᵧ).dot(Ⲏ)
		HΨH = (HΨ.T - np.mean(HΨ.T, axis=0)).T
		return HΨH

	def get_Kᵪ(X, Y, X_kernel, γ):
		if X_kernel == 'linear': 
			Kᵪ = X.dot(X.T)
		elif X_kernel == 'Gaussian':
			Kᵪ = sklearn.metrics.pairwise.rbf_kernel(X, gamma=γ)

		return Kᵪ

	def get_Kᵧ(X, Y, X_kernel, γ):
		if Y_kernel == 'linear': 
			Yₒ = OneHotEncoder(categories='auto', sparse=False).fit_transform(Y)
			Kᵧ = Yₒ.dot(Yₒ.T)
		elif X_kernel == 'Gaussian':
			[γᵪ, γᵧ] = get_γ(X, Y, sigma_type)
			Kᵧ = sklearn.metrics.pairwise.rbf_kernel(Y, gamma=γ)

		return Kᵧ


	if len(X.shape) == 1: X = np.reshape(X, (X.size, 1))
	if len(Y.shape) == 1: Y = np.reshape(Y, (Y.size, 1))
	n = X.shape[0]
	[γᵪ, γᵧ] = get_γ(X, Y, sigma_type)

	Kᵪ = get_Kᵪ(X, Y, X_kernel, γᵪ)
	Kᵧ = get_Kᵧ(X, Y, X_kernel, γᵧ)

	HKᵪ = Kᵪ - np.mean(Kᵪ, axis=0)					# equivalent to		HKᵪ = H.dot(Kᵪ)
	HKᵧ = Kᵧ - np.mean(Kᵧ, axis=0)                  # equivalent to		HKᵧ = H.dot(Kᵧ)
	Hᵪᵧ= np.sum(HKᵪ.T*HKᵧ)							# same as HKᵪH = double_center(Kᵪ)
                                                    #		  Hᵪᵧ = np.sum(HKᵪH*Kᵧ)
	if not normalize_hsic: return Hᵪᵧ/(n*n)

	Hᵪ = np.linalg.norm(HKᵪ)						# equivalent to 	np.sqrt(np.sum(KᵪH*KᵪH))
	Hᵧ = np.linalg.norm(HKᵧ) 						# equivalent to 	np.sqrt(np.sum(KᵧH*KᵧH))
	H = Hᵪᵧ/( Hᵪ * Hᵧ )

	return H


def accuracy(Y, Ŷ):
	if type(Y).__name__ == 'Tensor': 
		Y = Y.cpu().numpy()

	if type(Ŷ).__name__ == 'Tensor': 
		Ŷ = Ŷ.cpu().numpy()


	if type(Y).__name__ != 'ndarray' or type(Ŷ).__name__ != 'ndarray': 
		print('Error: Y must be numpy array or Tensor')
		exit()

	return accuracy_score(Y, Ŷ)


if __name__ == '__main__':
	n = 300
	
	#	Perfect Linear Data
	dat = np.random.rand(n,1)
	plinear_data = np.hstack((dat,dat)) + 1
	df = pd.DataFrame(data=plinear_data, columns=["x", "y"])

	enc = KBinsDiscretizer(n_bins=10, encode='ordinal')
	XP_data_nmi = np.squeeze(enc.fit_transform(np.atleast_2d(plinear_data[:,0]).T))
	enc = KBinsDiscretizer(n_bins=10, encode='ordinal')
	YP_data_nmi = np.squeeze(enc.fit_transform(np.atleast_2d(plinear_data[:,1]).T))

	plinear_pc = np.round(pearsonr(plinear_data[:,0], plinear_data[:,1])[0], 2)
	plinear_nmi = np.round(normalized_mutual_info_score(XP_data_nmi, YP_data_nmi),2)
	plinear_hsic = np.round(ℍ(plinear_data[:,0], plinear_data[:,1], sigma_type='opt'),2)	
	plinear_pps = np.round(pps.score(df, "x", "y")['ppscore'],2)

	print('Linear Relationship:')
	print('\tCorrelation : ', plinear_pc)
	print('\tNMI : ', plinear_nmi)
	print('\tpps : ', plinear_pps)
	print('\tHSIC : ', plinear_hsic)


	#	Linear Data
	dat = np.random.rand(n,1)
	linear_data = np.hstack((dat,dat)) + 0.04*np.random.randn(n,2)
	df = pd.DataFrame(data=linear_data, columns=["x", "y"])

	enc = KBinsDiscretizer(n_bins=10, encode='ordinal')
	XL_data_nmi = np.squeeze(enc.fit_transform(np.atleast_2d(linear_data[:,0]).T))
	enc = KBinsDiscretizer(n_bins=10, encode='ordinal')
	YL_data_nmi = np.squeeze(enc.fit_transform(np.atleast_2d(linear_data[:,1]).T))

	linear_pc = np.round(pearsonr(linear_data[:,0], linear_data[:,1])[0], 2)
	linear_nmi = np.round(normalized_mutual_info_score(XL_data_nmi, YL_data_nmi),2)
	linear_hsic = np.round(ℍ(linear_data[:,0], linear_data[:,1], sigma_type='opt'),2)	
	linear_pps = np.round(pps.score(df, "x", "y")['ppscore'],2)

	print('Linear Relationship:')
	print('\tCorrelation : ', linear_pc)
	print('\tNMI : ', linear_nmi)
	print('\tpps : ', linear_pps)
	print('\tHSIC : ', linear_hsic)

	#	Sine Data
	dat_x = 9.3*np.random.rand(n,1)
	dat_y = np.sin(dat_x)
	sine_data = np.hstack((dat_x,dat_y)) + 0.06*np.random.randn(n,2)
	df = pd.DataFrame(data=sine_data, columns=["x", "y"])
	sine_pc = np.round(pearsonr(sine_data[:,0], sine_data[:,1])[0],2)

	enc = KBinsDiscretizer(n_bins=10, encode='ordinal')
	Xsine_data_nmi = np.squeeze(enc.fit_transform(np.atleast_2d(sine_data[:,0]).T))
	enc = KBinsDiscretizer(n_bins=10, encode='ordinal')
	Ysine_data_nmi = np.squeeze(enc.fit_transform(np.atleast_2d(sine_data[:,1]).T))

	sine_nmi = np.round(normalized_mutual_info_score(Xsine_data_nmi, Ysine_data_nmi),2)
	sine_hsic = np.round(ℍ(sine_data[:,0], sine_data[:,1], sigma_type='opt'),2)
	sine_pps = np.round(pps.score(df, "x", "y")['ppscore'],2)

	print('Sine Relationship:')
	print('\tCorrelation : ', sine_pc)
	print('\tNMI : ', sine_nmi)
	print('\tpps : ', sine_pps)
	print('\tHSIC : ', sine_hsic)


	#	Parabola Data
	dat_x = 4*np.random.rand(n,1) - 2
	dat_y = 0.05*dat_x*dat_x
	para_data = np.hstack((dat_x,dat_y)) + 0.01*np.random.randn(n,2)
	df = pd.DataFrame(data=para_data, columns=["x", "y"])

	enc = KBinsDiscretizer(n_bins=10, encode='ordinal')
	Xp_data_nmi = np.squeeze(enc.fit_transform(np.atleast_2d(para_data[:,0]).T))
	enc = KBinsDiscretizer(n_bins=10, encode='ordinal')
	Yp_data_nmi = np.squeeze(enc.fit_transform(np.atleast_2d(para_data[:,1]).T))

	para_pc = np.round(pearsonr(para_data[:,0], para_data[:,1])[0],2)
	para_nmi = np.round(normalized_mutual_info_score(Xp_data_nmi, Yp_data_nmi),2)
	para_hsic = np.round(ℍ(para_data[:,0], para_data[:,1], sigma_type='opt'),2)
	para_pps = np.round(pps.score(df, "x", "y")['ppscore'],2)
	
	print('Parabola Relationship:')
	print('\tCorrelation : ', para_pc)
	print('\tNMI : ', para_nmi)
	print('\tpps : ', para_pps)
	print('\tHSIC : ', para_hsic)

	#	Random uniform Data
	unif_data = np.random.rand(n,2)
	df = pd.DataFrame(data=unif_data, columns=["x", "y"])

	enc = KBinsDiscretizer(n_bins=10, encode='ordinal')
	Xr_data_nmi = np.squeeze(enc.fit_transform(np.atleast_2d(unif_data[:,0]).T))
	enc = KBinsDiscretizer(n_bins=10, encode='ordinal')
	Yr_data_nmi = np.squeeze(enc.fit_transform(np.atleast_2d(unif_data[:,1]).T))


	unif_pc = np.round(pearsonr(unif_data[:,0], unif_data[:,1])[0],2)
	unif_hsic = np.round(ℍ(unif_data[:,0], unif_data[:,1], sigma_type='opt'),2)
	unif_nmi = np.round(normalized_mutual_info_score(Xr_data_nmi, Yr_data_nmi),2)
	unif_pps = np.round(pps.score(df, "x", "y")['ppscore'],2)
	
	print('Random Relationship:')
	print('\tCorrelation : ', unif_pc)
	print('\tNMI : ', unif_nmi)
	print('\tpps : ', unif_pps)
	print('\tHSIC : ', unif_hsic)


	plt.figure(figsize=(13,3))

	plt.subplot(151)
	plt.plot(plinear_data[:,0], plinear_data[:,1], 'bx')
	plt.title('$\\rho$ : ' + str(plinear_pc) + ' , HSIC : ' + str(plinear_hsic) + '\npps : ' + str(plinear_pps) + ' , nmi : ' + str(plinear_nmi))

	plt.subplot(152)
	plt.plot(linear_data[:,0], linear_data[:,1], 'bx')
	plt.title('$\\rho$ : ' + str(linear_pc) + ' , HSIC : ' + str(linear_hsic) + '\npps : ' + str(linear_pps) + ' , nmi : ' + str(linear_nmi))
	
	plt.subplot(153)
	plt.plot(sine_data[:,0], sine_data[:,1], 'bx')
	plt.title('$\\rho$ : ' + str(sine_pc) + ' , HSIC : ' + str(sine_hsic) + '\npps : ' + str(sine_pps) + ' , nmi : ' + str(sine_nmi))

	plt.subplot(154)
	plt.plot(para_data[:,0], para_data[:,1], 'bx')
	plt.title('$\\rho$ : ' + str(para_pc) + ' , HSIC : ' + str(para_hsic) + '\npps : ' + str(para_pps) + ' , nmi : ' + str(para_nmi))

	plt.subplot(155)
	plt.plot(unif_data[:,0], unif_data[:,1], 'bx')
	plt.title('$\\rho$ : ' + str(unif_pc) + ' , HSIC : ' + str(unif_hsic) + '\npps : ' + str(unif_pps) + ' , nmi : ' + str(unif_nmi))

	plt.tight_layout()
	plt.show()





#!/usr/bin/env python

from numpy import genfromtxt
from wuml.IO import *
from wuml.data_stats import *
from wuml.data_loading import *
from wuml.type_check import *

from sklearn.model_selection import KFold
from sklearn import preprocessing
import pandas as pd
import numpy as np
import sys
import os

np.set_printoptions(precision=4)
np.set_printoptions(threshold=30)
np.set_printoptions(linewidth=300)
np.set_printoptions(suppress=True)
np.set_printoptions(threshold=sys.maxsize)

def remove_rows_with_too_much_missing_entries(dataFrame, threshold=0.6, newDataFramePath=''):
	'''
		If a row is kept if it has more than "threshold" percentage of normal data
		dataFrame is a pandas format, it can be converted this way
		pd.DataFrame(data=data[1:,1:],    # values
						index=data[1:,0],    # 1st column as index
						columns=data[0,1:])  # 1st row as the column names
	'''
	df = dataFrame.get_data_as('DataFrame')

	X = df.values
	n = X.shape[0]
	d = X.shape[1]

	pth = './results/Preprocessing/'
	ensure_path_exists('./results')
	ensure_path_exists(pth)

	#	Obtain a decimated dataframe
	limitPer = int(d * threshold)
	df_decimated = df.dropna(thresh=limitPer, axis=0)
	oldID = df['id'].values.tolist()
	newID = df_decimated['id'].values.tolist()
	removed_samples = set(oldID).difference(newID)

	#	Record the results
	output_str = 'Original dataFrame dimension : %d samples,  %d dimensions\n'%(n, d)
	output_str += 'Deciated dataFrame dimension : %d samples,  %d dimensions\n'%(df_decimated.values.shape[0], df_decimated.values.shape[1])
	output_str += 'Removed Rows missing at least these percentage of entries : %.3f\n'%(1-threshold)

#	#	Record the removed features 
	output_str += '\nID of Removed Rows\n'
	output_str += str(removed_samples)

	write_to(output_str, pth + 'row_decimation_info.txt')
	if newDataFramePath != '': df_decimated.to_csv(path_or_buf=newDataFramePath, index=False)

	df_decimated = wData(dataFrame=df_decimated)
	return df_decimated


def remove_columns_with_too_much_missing_entries(dataFrame, threshold=0.6, newDataFramePath=''):
	'''
		If a column is kept if it has more than "threshold" percentage of normal data
		dataFrame is a pandas format, it can be converted this way
		pd.DataFrame(data=data[1:,1:],    # values
						index=data[1:,0],    # 1st column as index
						columns=data[0,1:])  # 1st row as the column names
	'''
	df = dataFrame.get_data_as('DataFrame')

	X = df.values
	n = X.shape[0]
	d = X.shape[1]

	pth = './results/Preprocessing/'
	ensure_path_exists('./results')
	ensure_path_exists(pth)

	#	Obtain a decimated dataframe
	limitPer = int(n * threshold)
	df_decimated = df.dropna(thresh=limitPer, axis=1)
	oldColumns = df.columns.values.tolist()
	newColumns = df_decimated.columns.values.tolist()
	removed_columns = set(oldColumns).difference(newColumns)

	#	Record the results
	output_str = 'Original dataFrame dimension : %d samples,  %d dimensions\n'%(n, d)
	output_str += 'Deciated dataFrame dimension : %d samples,  %d dimensions\n'%(df_decimated.values.shape[0], df_decimated.values.shape[1])
	output_str += 'Removed Columns missing at least these percentage of entries : %.3f\n'%(1-threshold)

	#	Record the removed features 
	output_str += '\nRemoved Columns + Missing Percentage'
	max_width = len(max(removed_columns, key=len))
	for column in removed_columns:
		x = df[column].values
		missing_percentage = np.sum(np.isnan(x))/n
		output_str += (('\n\t%-' + str(max_width) + 's\t%.2f')%(column, missing_percentage))

	#	Record the retained features 
	output_str += '\n\nRetained Columns + Missing Percentage'
	for column in df_decimated:
		x = df_decimated[column].values
		missing_percentage = np.sum(np.isnan(x))/n
		output_str += (('\n\t%-' + str(max_width) + 's\t%.2f')%(column, missing_percentage))

	write_to(output_str, pth + 'column_decimation_info.txt')
	if newDataFramePath != '': df_decimated.to_csv(path_or_buf=newDataFramePath, index=False)

	df_decimated = wData(dataFrame=df_decimated)
	return df_decimated

def decimate_data_with_missing_entries(dataFrame, column_threshold=0.6, row_threshold=0.6,newDataFramePath=''):
	'''
		It will automatically remove rows and columns of a dataFrame with missing entries.
	'''
	df = dataFrame.get_data_as('DataFrame')
	dfSize = 'Data size:' + str(df.shape)

	mdp = np.array(identify_missing_data_per_feature(dataFrame))
	x = np.arange(1, len(mdp)+1)

	lp = lines(figsize=(10,5))
	lp.plot_line(x, mdp, 'Before Missing Percentage', 'Feature ID', 'Percentage Missing', 
					imgText=dfSize, subplot=121, ylim=[0,1], xTextLoc=0, yTextLoc=0.9)

	df = remove_columns_with_too_much_missing_entries(dataFrame, threshold=column_threshold)
	df_decimated = remove_rows_with_too_much_missing_entries(df, threshold=row_threshold, newDataFramePath=newDataFramePath)
	dfSize = 'Data size:' + str(df_decimated.shape)

	mdp = np.array(identify_missing_data_per_feature(df_decimated))
	x = np.arange(1, len(mdp)+1)

	lp.plot_line(x, mdp, 'After Missing Percentage', 'Feature ID', 'Percentage Missing', 
					imgText=dfSize, subplot=122, ylim=[0,1], xTextLoc=0, yTextLoc=0.9)
	lp.show()

	return df_decimated


def center_and_scale(wuData):
	X = wuData.get_data_as('ndarray')
	X = preprocessing.scale(X)
	wuData.df = pd.DataFrame(X)
	return wuData



def center_scale_with_missing_data(X, replace_nan_with_0=False): 
	'''
		For each column, find μ, σ while ignoring the entries that are zero. 
	'''
	d = X.shape[1]
	ignore_column_with_0_σ = []
	for i in range(d):
		x = X[:,i]
		ẋ = x[np.invert(np.isnan(x))]
		ẋ = ẋ - np.mean(ẋ)
		σ = np.std(ẋ)

		if σ < 0.00001:
			ignore_column_with_0_σ.append(i)
		else:
			X[np.invert(np.isnan(x)), i] = ẋ/σ

	for i in ignore_column_with_0_σ:
		X = np.delete(X, i , axis=1)	# delete column with σ=0

	if replace_nan_with_0:
		X = np.nan_to_num(X)

	return X, ignore_column_with_0_σ


def gen_10_fold_data(data_name, data_path='./data/'):

	xpath = data_path + data_name

	X = np.loadtxt(xpath + '.csv', delimiter=',', dtype=np.float64)			
	Y = np.loadtxt(xpath + '_label.csv', delimiter=',', dtype=np.int32)			

	fold_path = xpath + '/'
	if os.path.exists(fold_path): 
		pass
	else:
		os.mkdir(fold_path)


	#if os.path.exists(fold_path): 
	#	txt = input("There's already a 10 fold data for %s, generate a new set? (y/N)"%data_name)
	#	if txt == 'y':
	#		pass
	#	else:
	#		return
	#else:
	#	os.mkdir(fold_path)


	kf = KFold(n_splits=10, shuffle=True)
	kf.get_n_splits(X)
	loopObj = enumerate(kf.split(X))

	for count, data in loopObj:
		[train_index, test_index] = data

		X_train, X_test = X[train_index], X[test_index]
		Y_train, Y_test = Y[train_index], Y[test_index]
		
		np.savetxt( fold_path + data_name + '_' + str(count+1) + '.csv', X_train, delimiter=',', fmt='%.6f') 
		np.savetxt( fold_path + data_name + '_' + str(count+1) + '_label.csv', Y_train, delimiter=',', fmt='%d') 
		np.savetxt( fold_path + data_name + '_' + str(count+1) + '_test.csv', X_test, delimiter=',', fmt='%.6f') 
		np.savetxt( fold_path + data_name + '_' + str(count+1) + '_label_test.csv', Y_test, delimiter=',', fmt='%d') 

def rearrange_sample_to_same_class(X,Y):
	l = np.unique(Y)
	newX = np.empty((0, X.shape[1]))
	newY = np.empty((0))

	for i in l:
		indices = np.where(Y == i)[0]
		newX = np.vstack((newX, X[indices, :]))
		newY = np.hstack((newY, Y[indices]))

	return [newX, newY]


#	Relating all samples to the most likely one, with p(X1) > p(Xi) for all i
#	Given X1 X2 with p(X1)/p(X2)=2  the weight for X1 = 1, and X2 = 2
def get_likelihood_weight(data):
	X = ensure_numpy(data)
	
	Pₓ = wuml.KDE(X)
	logLike = Pₓ(X, return_log_likelihood=True)
	max_likely = np.max(logLike)
	ratios = np.exp(max_likely - logLike)
	return wuml.wData(X_npArray=ratios)

